import os
import urllib.request
from tqdm import tqdm
import shutil
from contextlib import contextmanager

class ProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)

def download_url(url, output_path='.'):
    with ProgressBar(unit='B', unit_scale=True, miniters=1, desc=url.split('/')[-1]) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)

def extract_archive(archive, extract_path='.', delete=False):
    shutil.unpack_archive(archive, extract_path)
    if delete:
        os.remove(archive)


@contextmanager
def data_directory(dir):
    '''
    Changes context such that data directory is used as current work directory.
    Data directory is created if it does not exist.
    '''
    current_dir = os.getcwd()
    if not os.path.exists(dir):
        os.makedirs(dir)
    os.chdir(dir)
    try:
        yield
    finally:
        os.chdir(current_dir)


def kaggle_download(command):
    # TODO: add a check that kaggle is installed
    # TODO: add kaggle as package requirements
    os.system(f"kaggle {command}")



import numpy as np
import pandas as pd
from typing import List, Tuple
from .lcg import Lcg


class BalancedSplit():
    """Base class for splitting datasets into training and testing sets.

    Implements methods from [this paper](https://arxiv.org/abs/2211.10307).
    Its subclasses need to implement the `split` method.
    It should perform balanced splits separately for all classes.
    Its children are `IdentitySplit` and `TimeAwareSplit`.
    `IdentitySplit` has children `ClosedSetSplit`, `OpenSetSplit` and `DisjointSetSplit`.
    `TimeAwareSplit` has children `TimeProportionSplit` and `TimeCutoffSplit`.

    Attributes:
      df (pd.DataFrame): A dataframe of the data. It must contain columns
        `identity` for all splits and `date` for time-aware splits.
      lcg (Lcg): Random number generator LCG.
      n (int): Number of samples.
      n_class (int): Number of unique identities.
      y (np.ndarray): List of identities.
      y_counts (np.ndarray): List of sample counts for each unique identity.
      y_unique (np.ndarray): List of unique sorted identities.
    """

    def __init__(
            self,
            df: pd.DataFrame,
            seed: int = 666,
            identity_skip: str = 'unknown'
            ) -> None:
        """Initializes the class.

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns
                `identity` for all splits and `date` for time-aware splits.
            seed (int, optional): Initial seed for the LCG random generator.
            identity_skip (str, optional): Name of the identities to ignore.
        """

        # Potentially remove the unknown identities
        self.df = df.copy()
        self.df = self.df[self.df['identity'] != identity_skip]
        # Initialize the random number generator
        self.set_seed(seed)

    def set_seed(self, seed: int) -> None:
        """Changes the seed of the random number generator.

        Args:
            seed (int): The desired seed.
        """
        
        self.lcg = Lcg(seed)
    
    def split(self, *args, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """Splitting method which needs to be implemented by subclasses.

        It splits the dataframe `self.df` into labels `idx_train` and `idx_test`.
        The subdataset is obtained by `self.df.loc[idx_train]` (not `iloc`).

        Returns:
            List of labels of the training and testing sets.
        """

        raise(NotImplementedError('Subclasses should implement this. \n You may want to use ClosedSetSplit instead of BalancedSplit.'))

    def general_split(
            self,
            ratio_train: float,
            individual_train: List[str],
            individual_test: List[str],
            ) -> Tuple[np.ndarray, np.ndarray]:
        """General-purpose split into the training and testing sets.

        It puts all samples of `individual_train` into the training set
        and all samples of `individual_test` into the testing set.
        The splitting is performed for each individual separately.
        The split will result in at least one sample in both the training and testing sets.
        If only one sample is available for an individual, it will be in the training set.
                
        Args:
            ratio_train (float): *Approximate* size of the training set.
            individual_train (List[str]): Individuals to be only in the training test.
            individual_test (List[str]): Individuals to be only in the testing test.

        Returns:
            List of labels of the training and testing sets.
        """

        # Compute how many samples go automatically to the training and testing sets
        n_train = sum(self.y_counts[[k in individual_train for k in self.y_unique]])
        n_test = sum(self.y_counts[[k in individual_test for k in self.y_unique]])
        
        # Recompute ratio_train and adjust it to proper bounds 
        if n_train + n_test > 0 and n_train + n_test < self.n:
            ratio_train = (self.n*ratio_train - n_train) / (self.n - n_test - n_train)
        ratio_train = np.clip(ratio_train, 0, 1)
        
        idx_train = np.empty(self.n, dtype='bool')
        # Make a loop over all individuals
        for individual, y_count in zip(self.y_unique, self.y_counts):            
            if individual in individual_train and individual in individual_test:
                # Check if the class does not belong to both sets
                raise(Exception('Individual cannot be both in individual_train and individual_test.'))
            elif individual in individual_train:
                # Check if the class does not belong to the training set
                idx_train_class = np.ones(y_count, dtype='bool')
            elif individual in individual_test:
                # Check if the class does not belong to the testing set
                idx_train_class = np.zeros(y_count, dtype='bool')
            else:
                idx_train_class = np.zeros(y_count, dtype='bool')
                # Otherwise compute the number of samples in the training set
                n_train = np.round(ratio_train * y_count).astype(int)
                if n_train == y_count and n_train > 1:
                    n_train -= 1
                if n_train == 0:
                    n_train = 1
                # Create indices to the training set and randomly permute them                
                idx_permutation = self.lcg.random_permutation(y_count)
                idx_train_class[:n_train] = True                
                idx_train_class = idx_train_class[idx_permutation]
            # Save the indices
            idx_train[self.y == individual] = idx_train_class
        return np.array(self.df.index.values)[idx_train], np.array(self.df.index.values)[~idx_train]




import os
import shutil
from . import utils

class Downloader():
    def __init__(self, name=None):
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name

    def get_data(self, root, *args, **kwargs):
        self.download_print(os.path.join(root, self.name), *args, **kwargs)
        self.extract_print(os.path.join(root, self.name), *args, **kwargs)
        print('DATASET %s: FINISHED. If mass downloading, you can remove it from the list.' % self.name)
        print('')
    
    def download_print(self, root, *args, **kwargs):
        print('DATASET %s: DOWNLOADING STARTED.' % self.name)
        self.download(root, *args, **kwargs)

    def extract_print(self, root, *args, **kwargs):
        print('DATASET %s: EXTRACTING STARTED.' % self.name)
        self.extract(root, *args, **kwargs)

    def download(self, *args, **kwargs):
        raise NotImplemented('Needs to be implemented by subclasses.')
    
    def extract(self, *args, **kwargs):
        raise NotImplemented('Needs to be implemented by subclasses.')



class AAUZebraFish(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            utils.kaggle_download(f"datasets download -d 'aalborguniversity/aau-zebrafish-reid' --unzip")

    def extract(self, root):
        pass

class AerialCattle2017(Downloader):
    url = 'https://data.bris.ac.uk/datasets/tar/3owflku95bxsx24643cybxu3qh.zip'
    archive = '3owflku95bxsx24643cybxu3qh.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class ATRW(Downloader):
    url = 'https://github.com/cvwc2019/ATRWEvalScript/archive/refs/heads/main.zip'
    archive = 'main.zip'
    downloads = [
        # Wild dataset (Detection)
        ('https://lilablobssc.blob.core.windows.net/cvwc2019/test/atrw_detection_test.tar.gz', 'atrw_detection_test.tar.gz'),

        # Re-ID dataset
        ('https://lilablobssc.blob.core.windows.net/cvwc2019/train/atrw_reid_train.tar.gz', 'atrw_reid_train.tar.gz'),
        ('https://lilablobssc.blob.core.windows.net/cvwc2019/train/atrw_anno_reid_train.tar.gz', 'atrw_anno_reid_train.tar.gz'),
        ('https://lilablobssc.blob.core.windows.net/cvwc2019/test/atrw_reid_test.tar.gz', 'atrw_reid_test.tar.gz'),
        ('https://lilablobssc.blob.core.windows.net/cvwc2019/test/atrw_anno_reid_test.tar.gz', 'atrw_anno_reid_test.tar.gz'),
    ]

    def download(self, root):
        with utils.data_directory(root):
            for url, archive in self.downloads:
                utils.download_url(url, archive)

            # Evaluation scripts
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            for url, archive in self.downloads:
                archive_name = archive.split('.')[0]
                utils.extract_archive(archive, archive_name, delete=True)

            # Evaluation scripts
            utils.extract_archive(self.archive, 'eval_script', delete=True)

class BelugaID(Downloader):
    url = 'https://lilablobssc.blob.core.windows.net/liladata/wild-me/beluga.coco.tar.gz'
    archive = 'beluga.coco.tar.gz'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class BirdIndividualID(Downloader):
    url = 'https://drive.google.com/uc?id=1YT4w8yF44D-y9kdzgF38z2uYbHfpiDOA'
    archive = 'ferreira_et_al_2020.zip'

    def download(self, root):
        with utils.data_directory(root):
            import gdown
            exception_text = '''Dataset must be downloaded manually.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#birdindividualid'''
            try:
                gdown.download(self.url, self.archive, quiet=False)
            except:
                raise Exception(exception_text)
            if not os.path.exists(self.archive):
                raise Exception(exception_text)
            
    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

            # Create new folder for segmented images
            folder_new = os.getcwd() + 'Segmented'
            if not os.path.exists(folder_new):
                os.makedirs(folder_new)

            # Move segmented images to new folder
            folder_move = 'Cropped_pictures'
            shutil.move(folder_move, os.path.join(folder_new, folder_move))

class CTai(Downloader):
    url = 'https://github.com/cvjena/chimpanzee_faces/archive/refs/heads/master.zip'
    archive = 'master.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

            # Cleanup            
            shutil.rmtree('chimpanzee_faces-master/datasets_cropped_chimpanzee_faces/data_CZoo')

class CZoo(Downloader):
    url = 'https://github.com/cvjena/chimpanzee_faces/archive/refs/heads/master.zip'
    archive = 'master.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

            # Cleanup
            shutil.rmtree('chimpanzee_faces-master/datasets_cropped_chimpanzee_faces/data_CTai')

class Cows2021(Downloader):
    url = 'https://data.bris.ac.uk/datasets/tar/4vnrca7qw1642qlwxjadp87h7.zip'
    archive = '4vnrca7qw1642qlwxjadp87h7.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class Drosophila(Downloader):
    downloads = [
        ('https://dataverse.scholarsportal.info/api/access/datafile/71066', 'week1_Day1_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71067', 'week1_Day1_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71068', 'week1_Day1_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71069', 'week1_Day1_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71065', 'week1_Day1_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71071', 'week1_Day2_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71072', 'week1_Day2_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71073', 'week1_Day2_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71075', 'week1_Day2_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71070', 'week1_Day2_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71077', 'week1_Day3_01to04.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71078', 'week1_Day3_05to08.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71079', 'week1_Day3_09to12.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71080', 'week1_Day3_13to16.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71081', 'week1_Day3_17to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71083', 'week2_Day1_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71084', 'week2_Day1_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71085', 'week2_Day1_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71086', 'week2_Day1_train_16to20.zip'),        
        ('https://dataverse.scholarsportal.info/api/access/datafile/71082', 'week2_Day1_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71094', 'week2_Day2_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71095', 'week2_Day2_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71109', 'week2_Day2_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71110', 'week2_Day2_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71093', 'week2_Day2_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71111', 'week2_Day3_01to04.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71112', 'week2_Day3_05to08.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71115', 'week2_Day3_09to12.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71117', 'week2_Day3_13to16.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71118', 'week2_Day3_17to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71119', 'week3_Day1_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71120', 'week3_Day1_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71121', 'week3_Day1_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71124', 'week3_Day1_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71097', 'week3_Day1_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71125', 'week3_Day2_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71126', 'week3_Day2_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71127', 'week3_Day2_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71128', 'week3_Day2_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71107', 'week3_Day2_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71129', 'week3_Day3_01to04.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71130', 'week3_Day3_05to08.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71131', 'week3_Day3_09to12.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71132', 'week3_Day3_13to16.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71133', 'week3_Day3_17to20.zip'),
    ]

    def download(self, root):
        with utils.data_directory(root):
            for url, archive in self.downloads:
                utils.download_url(url, archive)

            # Download README
            url = 'https://dataverse.scholarsportal.info/api/access/datafile/71134'
            file = 'ReadMe_Drosophila.pdf'
            utils.download_url(url, file)

    def extract(self, root):
        with utils.data_directory(root):
            for url, archive in self.downloads:
                utils.extract_archive(archive, extract_path=os.path.splitext(archive)[0], delete=True)

class FriesianCattle2015(Downloader):
    url = 'https://data.bris.ac.uk/datasets/wurzq71kfm561ljahbwjhx9n3/wurzq71kfm561ljahbwjhx9n3.zip'
    archive = 'wurzq71kfm561ljahbwjhx9n3.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class FriesianCattle2017(Downloader):
    url = 'https://data.bris.ac.uk/datasets/2yizcfbkuv4352pzc32n54371r/2yizcfbkuv4352pzc32n54371r.zip'
    archive = '2yizcfbkuv4352pzc32n54371r.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class GiraffeZebraID(Downloader):
    url = 'https://lilablobssc.blob.core.windows.net/giraffe-zebra-id/gzgc.coco.tar.gz'
    archive = 'gzgc.coco.tar.gz'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class Giraffes(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            url = 'ftp://pbil.univ-lyon1.fr/pub/datasets/miele2021/'
            os.system(f"wget -rpk -l 10 -np -c --random-wait -U Mozilla {url} -P '.' ")

    def extract(self, root):
        pass

class HappyWhale(Downloader):
    archive = 'happy-whale-and-dolphin.zip'

    def download(self, root):
        with utils.data_directory(root):
            exception_text = '''Kaggle terms must be agreed with.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#happywhale'''
            try:
                utils.kaggle_download(f"competitions download -c happy-whale-and-dolphin")
            except:
                raise Exception(exception_text)
            if not os.path.exists(self.archive):
                raise Exception(exception_text)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class HumpbackWhaleID(Downloader):
    archive = 'humpback-whale-identification.zip'

    def download(self, root):
        with utils.data_directory(root):
            exception_text = '''Kaggle terms must be agreed with.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#humpbackwhale'''
            try:
                utils.kaggle_download(f"competitions download -c humpback-whale-identification")
            except:
                raise Exception(exception_text)
            if not os.path.exists(self.archive):
                raise Exception(exception_text)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class HyenaID2022(Downloader):
    url = 'https://lilablobssc.blob.core.windows.net/liladata/wild-me/hyena.coco.tar.gz'
    archive = 'hyena.coco.tar.gz'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class IPanda50(Downloader):
    downloads = [
        ('https://drive.google.com/uc?id=1nkh-g6a8JvWy-XsMaZqrN2AXoPlaXuFg', 'iPanda50-images.zip'),
        ('https://drive.google.com/uc?id=1gVREtFWkNec4xwqOyKkpuIQIyWU_Y_Ob', 'iPanda50-split.zip'),
        ('https://drive.google.com/uc?id=1jdACN98uOxedZDT-6X3rpbooLAAUEbNY', 'iPanda50-eyes-labels.zip'),
    ]

    def download(self, root):
        import gdown
        with utils.data_directory(root):
            for url, archive in self.downloads:
                gdown.download(url, archive, quiet=False)

    def extract(self, root):
        with utils.data_directory(root):
            for url, archive in self.downloads:
                utils.extract_archive(archive, delete=True)

class LeopardID2022(Downloader):
    url = 'https://lilablobssc.blob.core.windows.net/liladata/wild-me/leopard.coco.tar.gz'
    archive = 'leopard.coco.tar.gz'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class LionData(Downloader):
    url = 'https://github.com/tvanzyl/wildlife_reidentification/archive/refs/heads/main.zip'
    archive = 'main.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

            # Cleanup
            shutil.rmtree('wildlife_reidentification-main/Nyala_Data_Zero')

class MacaqueFaces(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            downloads = [
                ('https://github.com/clwitham/MacaqueFaces/raw/master/ModelSet/MacaqueFaces.zip', 'MacaqueFaces.zip'),
                ('https://github.com/clwitham/MacaqueFaces/raw/master/ModelSet/MacaqueFaces_ImageInfo.csv', 'MacaqueFaces_ImageInfo.csv'),
            ]
            for url, file in downloads:
                utils.download_url(url, file)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive('MacaqueFaces.zip', delete=True)

class NDD20(Downloader):
    url = 'https://data.ncl.ac.uk/ndownloader/files/22774175'
    archive = 'NDD20.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)    

class NOAARightWhale(Downloader):
    archive = 'noaa-right-whale-recognition.zip'

    def download(self, root):
        with utils.data_directory(root):
            exception_text = '''Kaggle terms must be agreed with.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#noaarightwhale'''
            try:
                utils.kaggle_download(f"competitions download -c noaa-right-whale-recognition")
            except:
                raise Exception(exception_text)
            if not os.path.exists(self.archive):
                raise Exception(exception_text)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)
            utils.extract_archive('imgs.zip', delete=True)

            # Move misplaced image
            shutil.move('w_7489.jpg', 'imgs')
            os.remove('w_7489.jpg.zip')

class NyalaData(Downloader):
    url = 'https://github.com/tvanzyl/wildlife_reidentification/archive/refs/heads/main.zip'
    archive = 'main.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):        
            archive = 'main.zip'
            utils.extract_archive(self.archive, delete=True)

            # Cleanup
            shutil.rmtree('wildlife_reidentification-main/Lion_Data_Zero')

class OpenCows2020(Downloader):
    url = 'https://data.bris.ac.uk/datasets/tar/10m32xl88x2b61zlkkgz3fml17.zip'
    archive = '10m32xl88x2b61zlkkgz3fml17.zip'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class SeaTurtleIDHeads(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            utils.kaggle_download(f"datasets download -d 'wildlifedatasets/seaturtleidheads' --unzip")

    def extract(self, root):
        pass

class SeaTurtleID(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            utils.kaggle_download(f"datasets download -d 'wildlifedatasets/seaturtleid' --unzip")

    def extract(self, root):
        pass

class SealID(Downloader):
    archive = '22b5191e-f24b-4457-93d3-95797c900fc0_ui65zipk.zip'
    
    def download(self, root, url):
        if url == '':
            raise(Exception('URL must be provided for SealID.\nCheck https://wildlifedatasets.github.io/wildlife-datasets/downloads/#sealid'))
        with utils.data_directory(root):        
            utils.download_url(url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)
            utils.extract_archive(os.path.join('SealID', 'full images.zip'), delete=True)
            utils.extract_archive(os.path.join('SealID', 'patches.zip'), delete=True)
            
            # Create new folder for segmented images
            folder_new = os.getcwd() + 'Segmented'
            if not os.path.exists(folder_new):
                os.makedirs(folder_new)
            
            # Move segmented images to new folder
            folder_move = os.path.join('patches', 'segmented')
            shutil.move(folder_move, os.path.join(folder_new, folder_move))
            folder_move = os.path.join('full images', 'segmented_database')
            shutil.move(folder_move, os.path.join(folder_new, folder_move))
            folder_move = os.path.join('full images', 'segmented_query')
            shutil.move(folder_move, os.path.join(folder_new, folder_move))
            file_copy = os.path.join('patches', 'annotation.csv')
            shutil.copy(file_copy, os.path.join(folder_new, file_copy))
            file_copy = os.path.join('full images', 'annotation.csv')
            shutil.copy(file_copy, os.path.join(folder_new, file_copy))            

class SMALST(Downloader):
    url = 'https://drive.google.com/uc?id=1yVy4--M4CNfE5x9wUr1QBmAXEcWb6PWF'
    archive = 'zebra_training_set.zip'

    def download(self, root):
        import gdown
        with utils.data_directory(root):
            gdown.download(self.url, self.archive, quiet=False)

    def extract(self, root):
        with utils.data_directory(root):
            os.system('jar xvf ' + self.archive)
            os.remove(self.archive)
            shutil.rmtree(os.path.join('zebra_training_set', 'annotations'))
            shutil.rmtree(os.path.join('zebra_training_set', 'texmap'))
            shutil.rmtree(os.path.join('zebra_training_set', 'uvflow'))

class StripeSpotter(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            urls = [
                'https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/stripespotter/data-20110718.zip',
                'https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/stripespotter/data-20110718.z02',
                'https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/stripespotter/data-20110718.z01',
                ]
            for url in urls:
                os.system(f"wget -P '.' {url}")

    def extract(self, root):
        with utils.data_directory(root):
            # Extract
            os.system(f"zip -s- data-20110718.zip -O data-full.zip")
            os.system(f"unzip data-full.zip")

            # Cleanup
            os.remove('data-20110718.zip')
            os.remove('data-20110718.z01')
            os.remove('data-20110718.z02')
            os.remove('data-full.zip')

class WhaleSharkID(Downloader):
    url = 'https://lilablobssc.blob.core.windows.net/whale-shark-id/whaleshark.coco.tar.gz'
    archive = 'whaleshark.coco.tar.gz'

    def download(self, root):
        with utils.data_directory(root):
            utils.download_url(self.url, self.archive)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)

class WNIGiraffes(Downloader):
    url = "https://lilablobssc.blob.core.windows.net/wni-giraffes/wni_giraffes_train_images.zip"
    archive = 'wni_giraffes_train_images.zip'
    url2 = 'https://lilablobssc.blob.core.windows.net/wni-giraffes/wni_giraffes_train.zip'
    archive2 = 'wni_giraffes_train.zip'

    def download(self, root):
        with utils.data_directory(root):
            exception_text = '''Dataset must be downloaded manually.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#wnigiraffes'''
            raise Exception(exception_text)
            #os.system(f'azcopy cp {self.url} {self.archive}')
            #utils.download_url(self.url2, self.archive2)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive(self.archive, delete=True)
            utils.extract_archive(self.archive2, delete=True)

class ZindiTurtleRecall(Downloader):
    def download(self, root):
        with utils.data_directory(root):
            downloads = [
                ('https://storage.googleapis.com/dm-turtle-recall/train.csv', 'train.csv'),
                ('https://storage.googleapis.com/dm-turtle-recall/extra_images.csv', 'extra_images.csv'),
                ('https://storage.googleapis.com/dm-turtle-recall/test.csv', 'test.csv'),
                ('https://storage.googleapis.com/dm-turtle-recall/images.tar', 'images.tar'),
            ]
            for url, file in downloads:
                utils.download_url(url, file)

    def extract(self, root):
        with utils.data_directory(root):
            utils.extract_archive('images.tar', 'images', delete=True)


class Segmented(Downloader):
    def get_data(self, root):
        with utils.data_directory(root):
            print("You are trying to download a segmented dataset %s. \
                    It is already included in its non-segmented version download. \
                    Skipping." % os.path.split(root)[1])




import os
import cv2
import numpy as np
import pandas as pd
from typing import List
from matplotlib import pyplot as plt
from PIL import Image
from ..datasets import utils

def get_image(path: str) -> Image:
    """Loads an image.

    Args:
        path (str): Path of the image.

    Returns:
        Loaded image.
    """

    # We load it with OpenCV because PIL does not apply metadata.
    img = cv2.imread(path)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return Image.fromarray(img)

def plot_image(img: Image) -> None:
    """Plots an image.

    Args:
        img (Image): Image to be plotted.
    """
    
    fig, ax = plt.subplots()
    ax.imshow(img)
    plt.show()
    
def plot_segmentation(img: Image, segmentation: List[float]) -> None:
    """Plots an image and its segmentation mask.

    Args:
        img (Image): Image to be plotted.
        segmentation (List[float]): Segmentation mask in the form [x1, y1, x2, y2, ...].
    """

    if not np.isnan(segmentation).all():
        fig, ax = plt.subplots()
        ax.imshow(img)
        ax.plot(segmentation[0::2], segmentation[1::2], '--', linewidth=5, color='firebrick')
        plt.show()

def plot_bbox_segmentation(df: pd.DataFrame, root: str, n: int) -> None:
    """Plots n images from the dataframe `df`.

    If bounding boxes or segmentation masks are present, it plots them as well.

    Args:
        df (pd.DataFrame): Dataframe with column `path` (relative path)
            and possibly `bbox` and `segmentation`.
        root (str): Root folder where the images are stored.
        n (int): Number of images to plot.
    """

    if 'bbox' not in df.columns and 'segmentation' not in df.columns:
        for i in range(min(n, len(df))):
            img = get_image(os.path.join(root, df['path'].iloc[i]))
            plot_image(img)
    if 'bbox' in df.columns:
        df_red = df[~df['bbox'].isnull()]
        for i in range(min(n, len(df_red))):
            img = get_image(os.path.join(root, df_red['path'].iloc[i]))
            segmentation = utils.bbox_segmentation(df_red['bbox'].iloc[i])
            plot_segmentation(img, segmentation)
    if 'segmentation' in df.columns:
        df_red = df[~df['segmentation'].isnull()]
        for i in range(min(n, len(df_red))):
            segmentation = df_red['segmentation'].iloc[i]
            if type(segmentation) == str:
                img = get_image(os.path.join(root, df_red['path'].iloc[i]))
                plot_image(img)
                img = get_image(os.path.join(root, segmentation))
                plot_image(img)
            elif type(segmentation) == list or type(segmentation) == np.ndarray:
                img = get_image(os.path.join(root, df_red['path'].iloc[i]))
                segmentation = df_red['segmentation'].iloc[i]
                plot_segmentation(img, segmentation)

def plot_grid(
        df: pd.DataFrame,
        root: str,
        n_rows: int = 5,
        n_cols: int = 8,
        offset: float = 10,
        img_min: float = 100,
        rotate: bool = True
        ) -> Image:
    """Plots a grid of size (n_rows, n_cols) with images from the dataframe.

    Args:
        df (pd.DataFrame): Dataframe with column `path` (relative path).
        root (str): Root folder where the images are stored. 
        n_rows (int, optional): The number of rows in the grid.
        n_cols (int, optional): The number of columns in the grid.
        offset (float, optional): The offset between images.
        img_min (float, optional): The minimal size of the plotted images.
        rotate (bool, optional): Rotates the images to have the same orientation.

    Returns:
        The plotted grid.
    """

    # Select indices of images to be plotted
    n = min(len(df), n_rows*n_cols)
    idx = np.random.permutation(len(df))[:n]

    # Load images and compute their ratio
    ratios = []
    for k in idx:
        file_path = os.path.join(root, df.iloc[k]['path'])
        im = get_image(file_path)
        ratios.append(im.size[0] / im.size[1])

    # Get the size of the images after being resized
    ratio = np.median(ratios)
    if ratio > 1:    
        img_w, img_h = int(img_min*ratio), int(img_min)
    else:
        img_w, img_h = int(img_min), int(img_min/ratio)

    # Create an empty image grid
    im_grid = Image.new('RGB', (n_cols*img_w + (n_cols-1)*offset, n_rows*img_h + (n_rows-1)*offset))

    # Fill the grid image by image
    for k in range(n):
        i = k // n_cols
        j = k % n_cols

        # Load the image
        file_path = os.path.join(root, df.iloc[idx[k]]['path'])
        im = get_image(file_path)

        # Possibly rotate the image
        if rotate and ((ratio > 1 and im.size[0] < im.size[1]) or (ratio < 1 and im.size[0] > im.size[1])):
            im = im.transpose(Image.ROTATE_90)

        # Rescale the image
        im.thumbnail((img_w,img_h))

        # Place the image on the grid
        pos_x = j*img_w + j*offset
        pos_y = i*img_h + i*offset        
        im_grid.paste(im, (pos_x,pos_y))
    return im_grid
